"""
Generator class for some Process-based soloing algorithms.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import random
import inspect
from pydub import AudioSegment
from GenerIter.process import Process
from GenerIter.util import debug
import GenerIter.excepts as robox


class Solo(Process):
    
    SOLOMAP = [
        {
            "upper" : 10000,
            "variation" : (800, 2400),
            "fades" : 100,
            "back_silence" : 4000
        },
        {
            "upper" : 60000,
            "variation" : (8000, 24000),
            "fades" : 3000,
            "front_silence" : 20000,
            "back_silence" : 40000
        },
        {
            "upper" : 60000,
            "variation" : (20000, 36000),
            "fades" : 3000,
            "front_silence" : 60000,
            "back_silence" : 40000
        },
        {
            "upper" : 10000,
            "variation" : (200, 800),
            "fades" : 10,
            "front_silence" : 60000,
            "back_silence" : 40000,
            "reps" : (5, 30)
        }
    ]
    
    def __init__(self):
        super().__init__()
        debug('Solo()')

    def generic(self):
        VOL_UP = 14
        VOL_DN = 18
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        
        voice = self._config["voice"]

        for ctr in range(iterations):
            # Grab some source material
            sample = self._inventory.selectRandom(voice)
            newAudio = AudioSegment.from_wav(sample)
            alen = len(newAudio)

            # Randomly select the parameters
            params = random.choice(self.SOLOMAP)

            # Are we trimming the input sample?
            if alen > params["upper"]:
                # Pick a random length
                samplen = random.randrange(params["variation"][0], params["variation"][1])
                # Define the range in which we can start the cut
                sampst = alen - samplen
                # Generate a start cut point in the defined range
                t1 = random.randrange(sampst)
                # Derive the endpoint of the cut
                t2 = (t1 + samplen)
                # Cut
                newAudio = newAudio[t1:t2]
                
            # Set a volume
            newvol = random.randrange(VOL_UP, VOL_DN)
            newAudio = newAudio - newvol

            # Set fades as defined by the params
            newAudio = self.declick(newAudio, params["fades"])

            # Do we need to pad front and/or back with some silence?
            if "front_silence" in params:
                if "back_silence" in params:
                    fsil = random.randrange(params["front_silence"])
                    bsil = random.randrange(params["back_silence"])
                    front = AudioSegment.silent(duration=fsil)
                    back = AudioSegment.silent(duration=bsil)
                    newAudio = front + newAudio + back
                else:
                    fsil = random.randrange(params["front_silence"])
                    front = AudioSegment.silent(duration=fsil)
                    newAudio = front + newAudio
            else:
                if "back_silence" in params:
                    bsil = random.randrange(params["back_silence"])
                    back = AudioSegment.silent(duration=bsil)
                    newAudio = newAudio + back
            
            # Do we wish to build repeats of this sample?
            if "reps" in params:
                addAudio = newAudio
                reps = random.randrange(params["reps"][0], params["reps"][1])
                for sam in range(reps):
                    newAudio += addAudio

            # Randomise front and back padding on the whole shebang
            dic = random.randrange(10)
            if dic == 7:
                sil1 = random.randrange(90000)
                sil2 = random.randrange(90000)
                solofront = AudioSegment.silent(duration = sil1)
                soloback = AudioSegment.silent(duration = sil2)
                newAudio = solofront + newAudio + soloback

            # Write it out
            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=newAudio)


    def multivoice_serial_ordered(self):
        """ 
        """
        declick = 10
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        voices = self._config["voices"]
        nvoices = len(voices)

        for ctr in range(iterations):
            audios = []
            # Set a soft threshold for the size of the piece
            size_limit = self.threshold()
            solo_base = AudioSegment.empty()
            # Randomly select our samples
            for voice in voices:
                samp = self._inventory.selectRandom(voice)
                audio = self.getsegment(sample=samp, muted=0, fade=0)
                audios.append(audio)

            ctr2 = 0
            while solo_base.duration_seconds < size_limit:
                solo_base = solo_base + self.bracket(audios[ctr2 % nvoices])
                ctr2 += 1

            # Write it out
            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=solo_base)



"""
Classes to represent references to GenerIter source files.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import GenerIter.excepts as robox
from GenerIter.util import jStr, debug

class Source():
    """Generic file source representation.

    Args:
        path (string) : string representation of the absolute or relative path to the target file.
        exist (boolean) : flag setting to test for the file's existence when the reference object is instantiated.

    Raises:
        GIResourceErr : if ``exist == True`` and the file does not exist.
    """
    def __init__(self, path=None, exist=False):
        """Init docstring
        """
        self._path = None
        self._fname = None
        self._root = None
        self._ext = None
        
        self.path = path
        if exist is True:
            if self.exists() is not True:
                raise robox.GIResourceErr("File {0} does not exist.".format(self.path))

    @property
    def path(self):
        """Accessor to the  value of full path value
        """
        return self._path

    @path.setter
    def path(self, value):
        """Sets value of full path.
 
        Also derives various component features from the path
        """
        if value is None:
            raise robox.GIParameterErr("No path given on Source class path setting accessor.")
        self._path = value
        self._fname = os.path.basename(self._path)
        self._dname = os.path.basename(os.path.dirname(self._path))
        self._root, self._ext = os.path.splitext(self._fname)

    @property
    def root(self):
        """Returns the root name of the file.
        """
        return self._root

    @property
    def fname(self):
        """Returns the basename of the file.
        """
        return self._fname

    @property
    def dname(self):
        """Returns the name of the file's directory.
        """
        return self._dname

    @property
    def ext(self):
        """Returns the filename extension of the file
        """
        return self._ext

    def isValidExtension(self, ref=None):
        """Tests if the filename extension matches an expected value.

        Indicates whether the Source or derived object carries the same file extension, either as upper- or lower-case forms.

        Args:
            ref (str) : the extension value to compare against.

        Returns:
            boolean
        """
        if ref is None:
            # This is essentially a NOOP
            return True
        else:
            if self.ext is None:
                return False
            case1 = (ref == self.ext.upper())
            case2 = (ref == self.ext.lower())
            return (case1 | case2)

    def exists(self):
        """Test if the file referred to exists.

        Returns:
            boolean
        """
        return os.path.isfile(self.path)

    def __str__(self):
        """Returns a string representation of the object variables.
        """
        return jStr(vars(self))

class WavSource(Source):
    """Derived class specialised for WAV source files.
    """
    def __init__(self, dpath=None, dexist=False):
        """Instantiates as a Source file then validates the filename extension.
        """
        super(WavSource, self).__init__(path=dpath, exist=dexist)
        if self.isValidExtension('.wav') is not True:
            raise robox.GIValidationErr("Incorrect file extension for WavSource.")

class FlacSource(Source):
    """Derived class specialised for FLAC source files.
    """
    def __init__(self, dpath=None, dexist=False):
        """Instantiates as a Source file then validates the filename extension.
        """
        super(FlacSource, self).__init__(path=dpath, exist=dexist)
        if self.isValidExtension('.flac') is not True:
            raise robox.GIValidationErr("Incorrect file extension for FlacSource.")

class Mp3Source(Source):
    """Derived class specialised for MP3 source files.
    """
    def __init__(self, dpath=None, dexist=False):
        """Instantiates as a Source file then validates the filename extension.
        """
        super(Mp3Source, self).__init__(path=dpath, exist=dexist)
        if self.isValidExtension('.mp3') is not True:
            raise robox.GIValidationErr("Incorrect file extension for Mp3Source.")
        



"""
Class that constructs the correct processor object.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
from GenerIter.selector import Selector
from GenerIter.config import Config
import GenerIter.excepts as robox
from GenerIter.util import debug

class ProcessorFactory():

    def __init__(self, vname, pname, fname, procmodule='GenerIter.processor'):
        self._vname = vname
        self._pname = pname
        self._method = None
        self._klass = None
        
        if self._vname is not None:
            modspec = '{0}.{1}'.format(procmodule, self._vname)
            try:
                mod = __import__(modspec, fromlist=[self._vname])
            except ImportError:
                raise ImportError('Could not import {0}.{1}'.format(modspec, self._vname))
            # Load the class from the module
            klass = getattr(mod, self._vname)
            # Create an object instance
            self._klass = klass()
            if self._klass is None:
                raise ImportError('Failed to instantiate {0}.{1}.{1}'.format(modspec, self._vname))
            # If we got here, we have a class instantiation
            # Let's see if we can get a reference to a method
            try:
                print("{0}.{1}()".format(self._vname, self._pname))
                self._method = getattr(self._klass, self._pname)
            except AttributeError:
                raise AttributeError('Could not obtain reference to {0}.{1}'.format(self._name, self._pname))

    @property
    def klass(self):
        return self._klass

    @property
    def method(self):
        return self._method

    def setMethod(self, fname):
        self._method = None
        if self._klass is not None:
            try:
                self._method = getattr(self._klass, fname)
            except AttributeError:
                print('Could not obtain {0}.{1}'.format(self._name, fname))
                sys.exit(-1)
        return self._method

    def configure(self, invent, config, dest, form, size):
        self._klass.configure(inventory=invent, configuration=config, destination=dest, forrmat=form, tsize=size)

    def process(self):
        self._method()
    



"""
Class to configure generator algorithms.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""

import os
import json
import GenerIter.excepts as robox
from GenerIter.util import jStr, debug, debug_except

class Config():
    
    def __init__(self, confpath=None):
        self._data = {}
        if confpath is not None:
            self.load(inpath=confpath)

    def load(self, inpath):
        if inpath is not None:
            with open(inpath) as fp:
                self._data = json.load(fp)

    def subcats(self):
        retval = []
        for key in self._data:
            if key != "Globals":
                retval.append(key)
        return retval

    def __str__(self):
        return jStr(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, val):
        self._data[key] = val



"""
Abstract base class for Command Line Interface apps in the GenerIter ecosystem

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import GenerIter.excepts as gix

class CLIBase():

    def __init__(self):
        """
        Common core initialisation sequence for all CLIBase objects.

        First all the command line options and arguments are parsed and validated.
        Secondly, the application processing function is executed.
        """
        self.parseArguments()
        self.process()

    def parseArguments(self):
        """
        There is no implementation of this function for the abstract base class.

        Raises:
            GenerIter.excepts.GINotImplementedErr

        """
        raise gix.GINotImplementedErr("The method parseArguments() is not implemented on this object.")

    def process(self):
        """
        There is no implementation of this function for the abstract base class.

        Raises:
            GenerIter.excepts.GINotImplementedErr

        """
        raise gix.GINotImplementedErr("The method process() is not implemented on this object.")



"""
Generator class for some basic Process-based algorithms for rhythmic generation.
"""
import os
import random
import inspect
from pydub import AudioSegment
from GenerIter.process import Process
from GenerIter.util import debug, nextPowerOf2
import GenerIter.excepts as robox

class Basic(Process):
    def __init__(self):
        super().__init__()
        debug('Basic()')

    def beatsbassdrone(self):
        """ This is a developmental prototype variant of the algorithm01 in the original
        DJProcessor code.

        As the name suggests, it expects to find categories in the sample inventory mapped into the *Beats*, *Bass* and *Drone* voices. 
        Otherise this algorithm will fail.
        
        Warning:
            This is example code used during early development, and should really only be used for reference and study.

        """
        # These could be parameterised in the config
        declick = 10
        drums_vol_range = (14, 18)
        bass_vol_range = (18, 22)
        pads_vol_range = (18, 22)
        
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        # Need to be able to pad the correct number of zeroes
        #digits = self.intwidth(iterations)
        # What's the base root name of the outputs?
        #track = self._config["track"]
        # Where are we sending the outputs?
        #dest = self._destination
        # Track repeats
        repeats = int(self._config["repeats"])

        for ctr in range(iterations):
            # Randomly select our samples
            drums_sample = self._inventory.selectRandom("Beats")
            bass_sample = self._inventory.selectRandom("Bass")
            pads_sample = self._inventory.selectRandom("Drone")

            # Instantiate the audio segments
            drums_audio = self.getsegment(drums_sample, drums_vol_range, declick)
            bass_audio = self.getsegment(bass_sample, bass_vol_range, declick)
            pads_audio = self.getsegment(pads_sample, pads_vol_range, declick)

            bb_overlay = drums_audio.overlay(bass_audio)
            composite_01 = bb_overlay + bb_overlay + bb_overlay + bb_overlay

            dbb_overlay = composite_01.overlay(pads_audio)

            for ctr2 in range(repeats):
                dbb_overlay = dbb_overlay + dbb_overlay

            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=dbb_overlay)


    def voices3(self):
        """ This developmental prototype algorithm uses the same algorithm as beatsbassdrone, 
        but allows the user to designate the 3 voices to be used in the compose file.

        The number of voices allowed is hard-coded to 3 only.

        
        Warning:
            This is example code used during early development, and should really only be used for reference and study.

        """
        # These could be parameterised in the config
        declick = 10
        aa_vol_range = (14, 18)
        ba_vol_range = (18, 22)
        ca_vol_range = (18, 22)

        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        # Need to be able to pad the correct number of zeroes
        #digits = self.intwidth(iterations)
        # What's the base root name of the outputs?
        #track = self._config["track"]
        # Where are we sending the outputs?
        #dest = self._destination
        # Track repeats
        repeats = int(self._config["repeats"])

        voices = self._config["voices"]

        for ctr in range(iterations):
            # Set a soft threshold for the size of the piece
            size_limit = self.threshold()
            # Randomly select our samples
            aa_sample = self._inventory.selectRandom(voices[0])
            ba_sample = self._inventory.selectRandom(voices[1])
            ca_sample = self._inventory.selectRandom(voices[2])

            # Instantiate the audio segments
            aa_audio = self.getsegment(aa_sample, aa_vol_range, declick)
            ba_audio = self.getsegment(ba_sample, ba_vol_range, declick)
            ca_audio = self.getsegment(ca_sample, ca_vol_range, declick)

            bb_overlay = aa_audio.overlay(ba_audio)
            composite_01 = bb_overlay + bb_overlay + bb_overlay + bb_overlay

            dbb_overlay = composite_01.overlay(ca_audio)

            for ctr2 in range(repeats):
                dbb_overlay = dbb_overlay + dbb_overlay

            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=dbb_overlay)

    def voices(self):
        """ This developmental prototype algorithm uses the same algorithm as voices3, 
        but allows the user to designate the 3 or more voices to be used in the compose file.

        Warning:
            This is example code used during early development, and should really only be used for reference and study.

        """
        # These could be parameterised in the config
        declick = 10
        vol_range = (14, 22)

        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        # Track repeats
        repeats = int(self._config["repeats"])

        voices = self._config["voices"]
        nvoices = len(voices)
        if nvoices < 3:
            raise robox.GIParameterErr("Insufficient voices specified for the 'voices' algorithm")
        
        for ctr in range(iterations):
            audios = []
            # Randomly select our samples
            for voice in voices:
                sample = self._inventory.selectRandom(voice)
                audio = self.getsegment(sample, vol_range, declick)
                audios.append(audio)

            summation = audios[0]
            for nctr in range(nvoices-2):
                summation = summation.overlay(audios[nctr+1])

            composite = summation + summation + summation + summation

            final_overlay = composite.overlay(audios[nvoices-1])

            for ctr2 in range(repeats):
                final_overlay = final_overlay + final_overlay

            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=final_overlay)

    def voices_shifted(self):
        """ This algorithm uses the same algorithm as voices, 
        but implements a sequential shift as each new voice is incorporated.

        Warning:
            This is example code used during early development, and should really only be used for reference and study.
        """
        # These could be parameterised in the config
        declick = 10
        vol_range = (14, 22)

        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        # Track repeats
        repeats = int(self._config["repeats"])

        voices = self._config["voices"]
        nvoices = len(voices)
        if nvoices < 3:
            raise robox.GIParameterErr("Insufficient voices specified for the 'voices' algorithm")

        for ctr in range(iterations):
            audios = []
            # Randomly select our samples
            for voice in voices:
                sample = self._inventory.selectRandom(voice)
                audio = self.getsegment(sample, vol_range, declick)
                audios.append(audio)

            summation = audios[0]
            for nctr in range(nvoices-1):
                summation = summation.overlay(audios[nctr+1])
                summation = summation + summation

            for ctr2 in range(repeats):
                summation = summation + summation

            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=summation)


    def groove(self):
        """The `Basic.groove` algorithm is used to form a rhythmic backbone to a composition. This algorithm attempts to make sensible alignment
        choices for the samples in use. It is considered a *production* algorithm.

        .. image:: images/Basic.groove.jpg

        The number of voices can be 3 or more. In this example we are using 4, which we can think of as **A**, **B**, **C** and **D** corresponding
        to their declaration position in the configuration list.

        Segments **A** and **B** are randomly selected from their inventory categories. These are then length-aligned. That is, whichever is the
        shorter of the two is padded with exactly the right amount of silence to make the frame lengths of both segments equal. These are then overlayed
        to form a composite rhythmic unit referred to as **AB**.

        The basic rhythmic beat is created by replicating a sequence of these **AB** segments end-to-end until the track size limit is surpassed.
        This is the `groove_base`.

        The `cycle` parameter specifies the number of these **AB** components tahe represent a higher level repetition unit called the `cycle` (obviously).

        Subsequent voices are also length-aligned (or curtailed) to match the size of the **AB** unit. A new segment, `groove_layer` is created,
        into which the later voices **C**, **D**, etc. are injected in a cyclic sequence, aligned precisely to the `groove_base` cycles and beats.

        Finally, the `groove_base` is overlayed with the `groove_layer` to create a composite whole track which is output as an audio file.

        Parameters:
            tracks (int) : number of times the process is to run.
            voices (list) : list of voice categories to select and use in the track
            cycle (int) : length of a repeat cycle in terms of groove_base *beats*

        Example:

        .. code:: json

           {
               "Basic" : {
                   "groove" : {
                       "tracks" : 50,
                       "cycle" : 4,
                       "voices" : [
                           "Bass",
                           "Beat",
                           "Percussion",
                           "Pad"
                       ]
                    }
                }
            }

        Raises:
            GenerIter.excepts.GIParameterErr
        """
        declick = 10
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        voices = self._config["voices"]
        nvoices = len(voices)
        cycle = self._config["cycle"]
      
        if nvoices < 3:
            raise robox.GIParameterErr("Insufficient voices specified for the groove algorithm")

        # Need to keep all overlays below the clipping threshold
        mute = 6.0 * nextPowerOf2(nvoices)

        for ctr in range(iterations):
            audios = []
            maxlen = 0
            # Set a soft threshold for the size of the piece
            size_limit = self.threshold()
            # Randomly select our samples
            for voice in voices:
                sample = self._inventory.selectRandom(voice)
                audio = self.getsegment(sample, mute, declick)
                audios.append(audio)

            for aud in range(2):
                if len(audios[aud]) > maxlen:
                    maxlen = len(audios[aud])
                    
            for aud in range(2, nvoices):
                audios[aud] = self.padtolength(audios[aud], maxlen, declick)

            # Create the groove's rhythmic base layer
            rhythm = audios[0].overlay(audios[1])
            rlen = len(rhythm)
            groove_base = rhythm
            print("{0}  < {1}".format(str(groove_base.duration_seconds), str(size_limit)))
            while groove_base.duration_seconds < size_limit:
                #groove_base = groove_base + rhythm
                groove_base = groove_base + rhythm

            # Create a blank segment of same size and frame rate
            groove_layer = AudioSegment.silent(duration=len(groove_base),
                                               frame_rate=groove_base.frame_rate)

            for ctr2 in range(2, nvoices):
                compound = AudioSegment.silent(duration=cycle*len(rhythm),
                                               frame_rate=groove_layer.frame_rate)
                offset = ((ctr2 - 2) % cycle) * rlen
                compound = compound.overlay(audios[ctr2], position=offset)
                groove_layer = groove_layer.overlay(compound, loop=True)

            summation = groove_base.overlay(groove_layer)

            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=summation)



"""
Class to catalogue and select source files.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import string
import random
import json
from pathlib import Path
import GenerIter.excepts as robox
from GenerIter.source import WavSource
from GenerIter.config import Config
from GenerIter.util import jStr, debug, debug_except

class Selector():
    """Represents the inventory of samples available for algorithmic processing.
    
    .. code-block:: python
       :caption: Example constructor usage
       
       from GenerIter.selector import Selector
       # Selector
       selector = Selector(searchpath=pathstring,loadpath=loadfile)

    """
    
    def __init__(self, searchpath=None, loadpath=None):
        self._data = {}
        if searchpath is not None:
            self.search(spath=searchpath)
        if loadpath is not None:
            self.load(lpath=loadpath)

    def subcats(self):
        """Get the list of top-level sub-categories in the Selector.

        Returns:
            [] (str)
        .. code-block:: python
           :caption: Example usage
           
           # Enumerate the sub-categories
           cats = selector.subcats()
           for cat in cats:
               # Get the sub-category
               category = selector[cat]
        
        """
        retval = []
        for key in self._data:
            retval.append(key)
        return retval
        
    def __str__(self):
        return jStr(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def selectRandom(self, key):
        """Method for getting a random selection from within a sub-category of the Selector.

        This method will attempt to randomly choose an entry for the specified sub-category in the Selector.

        It will fail if there are no ``true`` enabled entries in the sub-category or if the randomised function repeatedly
        fails to find a ``true`` enabled entry because they are too sparse.

        The number of attempts is limited by the size of the sub-category array.

        Args:
            key (str) : the name of a sub-category within the Selector's structure.

        Raises:
            RDJParameterErr : if unable to select a return value.
        """
        retval = None
        if key is not None:
            # How many entries are there in this category?
            try:
                sellen = len(self._data[key])
            except KeyError:
                print("KeyError in Selector for key: {0}".format(key))
                print(jStr(self._data))
                raise KeyError
            # Let's not try selecting on an empty set
            if sellen > 0:
                # If I can't select after that number of tries, there's a probably an issue
                ctr = sellen
                # Set a limit
                while ctr > 0:
                    src, incl = random.choice(list(self._data[key].items()))
                    if incl is True:
                        # We have a selection
                        retval = src
                        # Skip to the end.
                        ctr = 0
                    else:
                        # Marked as excluded
                        retval = None
                    # Need to avoid an infinite loop if ALL the entries got
                    # accidentally excluded.
                    ctr = ctr - 1
        # Houston, we have a problem
        if retval is None:
            raise robox.RDJParameterErr("Unable to select a valid source for {0}".format(key))
        # Here's a result.
        print("{0} : {1}".format(key, retval))
        return retval

    def search(self, spath=None):
        """Walk a directory tree to add to the Selector configuration.

        This method walks the specified directory tree and adds any discovered WAV files to its inventory.
        This method is uniquely additive in that it can be run repeatedly across the different or the same trees.
        Uniqueness is enforced during this process, so any repeats are silently overwritten.
        Any files encountered that do not match the criteria for a WAV file are ignored.

        Args:
            spath (str) : path to the root of the searchble directory tree.

        .. code-block:: python
           :caption: Example usage
       
           from GenerIter.selector import Selector
           # Empty Selector
           selector = Selector()
           # Search a directory tree
           selector.search(spath=pathstring)

        """
        if spath is not None:
            # Let's use the object-oriented filesystem abstraction
            sdir = Path(spath)
            # Recursively walk the directory and subdirectories
            fyles = list(sdir.rglob('*.*'))
            for fyle in fyles:
                # Ensure the filepath is absolute
                filepath = str(fyle.resolve())
                #debug(filepath)
                # Attempt to add this item into the current configuration
                try:
                    # Is it a valid WavSource?
                    src = WavSource(dpath=filepath, dexist=True)
                except Exception as inst:
                    debug(filepath)
                    #debug_except(inst)
                    debug("skipping ...")
                    src = None # The insert will skip this
                self.insert(source=src, include=True)

    def load(self, lpath=None):
        """Load a previously-saved Selector configuration.

        This method is uniquely additive in that it can be run repeatedly and any repeats are silently overwritten.

        Args:
            lpath (str) : path to the loadable JSON file containing a saved Selector state.

        .. code-block:: python
           :caption: Example usage
       
           from GenerIter.selector import Selector
           # Empty Selector
           selector = Selector()
           # Load a previously-saved inventory file
           selector.load(lpath=pathstring)

        """
        if lpath is not None:
            # Open the specified JSON file and load it
            with open(lpath) as fp:
                #debug("Open {0}".format(lpath))
                tcand = json.load(fp)
            # Iterate down the categories
            for category in tcand:
                # Within each category, iterate down the items
                for item in tcand[category]:
                    # Attempt to add this item into the current configuration
                    try:
                        # Is it a valid WavSource?
                        src = WavSource(dpath=item, dexist=True)
                    except robox.RDJValidationErr as err:
                        debug(err)
                        debug("WavSource exception for {0}".format(item))
                        src = None # The insert will skip this
                    self.insert(source=src, include=tcand[category][item])

    def insert(self, source, include=True):
        """Attempt to insert a source into the Selector configuration
        """
        #debug(source)
        if source is not None:
            # Force case-insensitivity in the keys
            lower_key = source.dname.lower()
            # Capitalise the first letter
            key = string.capwords(lower_key)
            if key in self._data:
                # If this key already exists, that's fine
                pass
            else:
                # If the key doesn't exist, create the dictionary for its entries
                self._data[key] = {}
            # Insert the entry in the correct subcategory
            self._data[key][source.path] = include
            

    



#!/usr/bin/env python3
"""
App to catalogue and select source files into a configuration.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import argparse
from pathlib import Path

from GenerIter.app.clibase import CLIBase
from GenerIter.selector import Selector
import GenerIter.excepts as gix
from GenerIter.util import debug

class Algorithm(CLIBase):

    klass_template = '''
from GenerIter.process import Process

class {KLASS}(Process):

    def __init__(self):
        super().__init__()


'''

    alg_template = '''
    deg {ALG}(self):
        print("Executes")

'''

    def _init_(self):
        super(Algorithm, self).__init__()


    def parseArguments(self):
        # Set up positional and optional arguments
        parser = argparse.ArgumentParser()
        # Zero or more -I parameters are allowed
        parser.add_argument("-A", help="Algorithm name",
                            action='store', required=True)
        # Zero or more -I parameters are allowed
        parser.add_argument("-M", help="Module name",
                            action='store', required=True)
        # This is a mandatory output name parameter.
        parser.add_argument("-L", help="Library name",
                            action='store', required=True)
        
        # Parse the command line as supplied
        args = parser.parse_args()
        # Set up the member values according to the command line options
        self._library = args.L.lower()
        self._module = args.M.lower()
        self._klass = self._module.capitalize()
        self._algorithm = args.A.lower()

    def build_lib(self):
        lib = Path(self._library)
        lib.mkdir(parents=True, exist_ok=True)
        init = lib / '__init__.py'
        if init.exists() is False:
            init.touch()

    def build_class(self):
        prog = '{0}.py'.format(self._module)
        mod = Path(self._library) / prog
        if mod.exists() is False:
            mod.open(mode='w')
            mod.write_text(self.klass_template.format(KLASS=self._klass))
            mod.close()

    def build_method(self):
        prog = '{0}.py'.format(self._module)
        mod = Path(self._library) / prog
        mod.open(mode='a')
        mod.write_text(self.alg_template.format(ALG=self_algorithm))
        mod.close()
            
    def process(self):
        self.build_lib()
        self.build_class()
        self.build_method()



#!/usr/bin/env python3
"""
App to generate music.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import argparse

from GenerIter.app.clibase import CLIBase
from GenerIter.selector import Selector
from GenerIter.config import Config
from GenerIter.factory import ProcessorFactory
import GenerIter.excepts as gix
from GenerIter.util import debug, localTimestamp

class Generator(CLIBase):

    def _init_(self):
        super(CLIBase, self).__init__()


    def parseArguments(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("-I", help="Source for inclusion in searches",
                            action='append')
        parser.add_argument("-L", help="Source selection file for inclusion in loads",
                            action='append')
        parser.add_argument("-C", help="Configuration file for algorithms",
                            action='store', required=True)
        
        args = parser.parse_args()
        
        self._includes = args.I
        self._loads = args.L
        #self._outfile = "{0}.wav".format(args.o)
        self._confname = args.C

    def loadSelections(self):
        self._selector = Selector()
        if self._includes is not None:
            for srcdir in self._includes:
                self._selector.search(srcdir)
        if self._loads is not None:
            for load in self._loads:
                self._selector.load(load)

    def loadConfiguration(self):
        self._configuration = Config()
        self._configuration.load(self._confname)
        #debug(self._configuration)
        self._voices = self._configuration.subcats()
        #debug(self._voices)
        self._sequence = None
        self._destination = os.getcwd()
        self._format = "wav"
        self._tsize = "m"
        try:
            if "sequence" in self._configuration["Globals"]:
                if len(self._configuration["Globals"]["sequence"]) > 0:
                    self._sequence = self._configuration["Globals"]["sequence"]
        except KeyError:
            pass

        try:
            if "destination" in self._configuration["Globals"]:
                self._destination = self._configuration["Globals"]["destination"]
        except KeyError:
            pass
        
        try:
            if "format" in self._configuration["Globals"]:
                self._format = self._configuration["Globals"]["format"]
        except KeyError:
            pass

        try:
            if "tsize" in self._configuration["Globals"]:
                self._tsize = self._configuration["Globals"]["tsize"].lower()
        except KeyError:
            pass

        ts = localTimestamp()
        self._destination = os.path.join(self._destination, ts)
        #debug(self._destination)
        #debug(self._sequence)
        #debug(self._format)
                

    def process(self):
        self.loadSelections()
        self.loadConfiguration()
        if self._sequence is not None:
            for voice in self._sequence:
                debug(f"Voice : {voice}")
                for processor in self._configuration[voice]:
                    debug(f"Processor : {processor}")
                    factory = ProcessorFactory(voice, processor, self._configuration[voice][processor])
                    factory.configure(invent=self._selector,
                                      config=self._configuration[voice][processor],
                                      dest=self._destination,
                                      form=self._format,
                                      size=self._tsize)
                    factory.process()
        else:
            for voice in self._voices:
                if voice != "Globals":
                    for processor in self._configuration[voice]:
                        factory = ProcessorFactory(voice, processor, self._configuration[voice][processor])
                        factory.configure(invent=self._selector,
                                          config=self._configuration[voice][processor],
                                          dest=self._destination,
                                          form=self._format,
                                          size=self._tsize)
                        factory.process()

#if __name__ == '__main__':
#    app = Generator()



#!/usr/bin/env python3
"""
App to catalogue and select source files into a configuration.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import argparse

from GenerIter.app.clibase import CLIBase
from GenerIter.selector import Selector
import GenerIter.excepts as gix
from GenerIter.util import debug

class Inventory(CLIBase):

    def _init_(self):
        super(Inventory, self).__init__()


    def parseArguments(self):
        # Set up positional and optional arguments
        parser = argparse.ArgumentParser()
        # Zero or more -I parameters are allowed
        parser.add_argument("-I", help="Source for inclusion in searches",
                            action='append')
        # Zero or more -I parameters are allowed
        parser.add_argument("-L", help="Source for inclusion in loads",
                            action='append')
        # This is a mandatory output name parameter.
        parser.add_argument("-o", help="Output file root name",
                            action='store', required=True)
        
        # Parse the command line as supplied
        args = parser.parse_args()
        # Set up the member values according to the command line options
        self._includes = args.I # This is a list
        self._loads = args.L # This is a list
        self._outfile = "{0}.json".format(args.o)


    def process(self):
        sel=Selector()
        if self._includes is not None:
            # Iterate over the search paths first
            for srcdir in self._includes:
                sel.search(srcdir)
        if self._loads is not None:
            # Iterate over any load files next
            for load in self._loads:
                sel.load(load)
        # Write it all out as a JSON file
        with open(self._outfile, "w") as of:
            of.write(str(sel))

#if __name__ == '__main__':
#    app = Inventory()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="GenerIter",
    version="0.2.0-dev20",
    author="Thomas Jackson Park & Jeremy Pavier",
    author_email="generiter@gmx.com",
    description="A package for Generative Iterative music composition.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GridPresence/GenerIter",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=['pydub'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points = {
        'console_scripts': ['generinv=GenerIter.app.clep_inventory:main',
                            'genercat=GenerIter.app.clep_categorise:main',
                            'generiter=GenerIter.app.clep_generator:main',
                            'generalg=GenerIter.app.clep_algorithm:main'],
    }
)



"""
Domain-specific Exception classes for GenerIter


Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""

class GIErr(Exception):
    """
    Base class for domain-specific exceptions.
    """

class GIParameterErr(GIErr):
    """
    Class for propogating parameter errors.
    """

class GIResourceErr(GIErr):
    """
    Class for propogating resource errors.
    """

class GIValidationErr(GIErr):
    """
    Class for propogating validation errors.
    """

class GINotImplementedErr(GIErr):
    """
    Class for propogating not implemented errors.
    """
    



"""
Generator class for some Process-based mixing algorithms.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import random
import inspect
from pydub import AudioSegment
from GenerIter.process import Process
from GenerIter.util import debug, nextPowerOf2
import GenerIter.excepts as robox

class Mix(Process):
    def __init__(self):
        super().__init__()
        debug('Mix()')

    def multitrack(self):
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        if "trackfade" in self._config:
            trackfade = self._config["trackfade"]
        else:
            trackfade = 0
        
        voices = self._config["voices"]
        nvoices = len(voices)

        if nvoices < 2:
            raise robox.GIParameterErr("Insufficient voices specified for the multitrack algorithm")

        mute = 6.0 * nextPowerOf2(nvoices)

        for ctr in range(iterations):
            audios = []
            # Randomly select our samples
            for voice in voices:
                level = voices[voice]
                sample = self._inventory.selectRandom(voice)
                audio = AudioSegment.from_wav(sample)
                #audio = audio.normalize()
                audio = audio + level
                audios.append(audio)

            summation = audios[0]
            for nctr in range(nvoices-1):
                summation = summation.overlay(audios[nctr+1], loop=True)

            if trackfade > 0:
                summation = self.declick(summation, trackfade)
            
            fname = inspect.currentframe().f_code.co_name
            self.write(algorithm=fname, counter=ctr, source=summation)



"""
Useful unencapsulated functions to be reused across the domain.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
from __future__ import print_function

import os
import errno
import json
from datetime import datetime, date
from subprocess import check_output, CalledProcessError

def debug(astring):
    """Only print if in DEBUG mode.

    Args:
        astring(str) : any valid Python string
    """
    if __debug__:
        print(astring)

def debug_except(inst):
    """Only print exception diagnostics if in DEBUG mode.

    Args:
        inst : any valid exception instance
    """
    if __debug__:
        print(type(inst))
        print(inst.args)
        print(inst)

def jsonSerial(obj):
    """JSON serializer for objects not serializable by default json code.

    Currently supports datetime objects.

    Args:
        obj (any type) : arbitary Python object or type.

    Returns:
        string

    Raises:
        TypeError : if the object is not serializable.
    """

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type {0} not serializable".format(type(obj)))

def jStr(struct):
    """Default JSON output human-readable string format.

    The output string is formatted for ease of reading, with an indent value of 4 chars,
    using standard separators and all fields sorted by name at their appropriate level.

    Args:
        struct : arbitary Python iterable data structure i.e. list or dict.

    Returns:
        string
    """
    if struct is None:
        struct = {}
    return json.dumps(struct,
                      default=jsonSerial,
                      sort_keys=True,
                      indent=4,
                      separators=(",", ":"))

def shCmd(cspec, trace=False):
    """Executes a shell command and returns a string list of the output.

    Args:
        cspec [] : array of command line options and parameters.
        trace (boolean) : flag to set for text output if required (default : False)

    Returns:
        [] (str) if ``trace == True`` else ``None``

    Raises:
        CalledProcessError : if the subprocess call fails
    """
    debug("Executing: {0}".format(cspec))
    retval = None
    try:
        intermed = check_output(cspec).decode("utf-8").rstrip()
        retval = intermed.split("\n")
        if trace is True:
            for line in retval:
                print(line)
    except CalledProcessError:
        retval = None
    return retval

def utf8(array):
    """Preserves byte strings, converts Unicode into UTF-8.

    Args:
        array (bytearray or str) : input array of bytes or chars

    Returns:
        UTF-8 encoded bytearray
    """
    retval = None
    if isinstance(array, bytes) is True:
        # No need to convert in this case
        retval = array
    else:
        retval = array.encode("utf-8")
    return retval


def mkdir_p(path):
    """Replicates mkdir -p functionality.

    For a given path, any missing directories are created to ensure the full path exists

    Args:
        path (str): absolute path to the target directory.

    Raises:
        OSError : if the path already exists as a file, or the target directory cannot be created because of a permissions error.
    """
    try:
        # Try to make all the necessary missing directories in the given path
        os.makedirs(path)
    except OSError as exc:
        # If the directory already exists on this path, that's okay
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        # Otherwise, it's either just a file or there's a permissions problem
        else:
            raise

def localTimestamp(some_time=None, time_format=None):
    """Return a specified UTC time, formatted by the given string.

    Args:
        some_time: a datetime object (default ``None`` uses ``datetime.utcnow()``)
        time_format: format to covert a datetime object to string (default ``None`` uses ``%Y%m%d%H%M%S``)

    Returns:
        A timestamp string
    """
    TIME_FORMAT_STRING = "%Y_%m%b_%dT%H_%M_%S_%fUTC"
    COMPACT_TIME_FMT = "%Y%m%d%H%M%S"
    if some_time is None:
        the_time = datetime.utcnow()
    else:
        the_time = some_time
    if time_format is None:
        the_time_format = COMPACT_TIME_FMT
    else:
        the_time_format = time_format
    time_string = the_time.strftime(the_time_format)
    return time_string


def nextPowerOf2(x):  
    return 1 if x == 0 else 2**(x - 1).bit_length()



#!/usr/bin/env python3
"""
App that is used to subcategorise a sample set according to a string search parameter.

The app will iterate through all the files in the current directory looking for files containing the category search pattern substring (the **-C** argument).
       
If the category search pattern is found in a filename, the file is moved into a subdirectory, which will be created if necessary. The 
name of the subdirectory will either correspond to the category search string or is specified by the **-D** argument.

Thus:

    .. code:: bash

       genercat -C Drums
        
and

    .. code:: bash

       genercat -C Drums -D Drums

are functionally identical, whereas

    .. code:: bash
        
       genercat -C Drums - D Percussion

puts the same files into a different subdirectory.

"""
import argparse
import pathlib

from GenerIter.app.clibase import CLIBase
import GenerIter.excepts as gix
from GenerIter.util import debug, mkdir_p

class Categorise(CLIBase):

    def _init_(self):
        super(Categorise, self).__init__()
        self._cat = None
        self._dest = None


    def parseArguments(self):
        """Parses the CLI arguments for the app.

        .. code:: bash

           usage: genercat [-h] -C C [-D D]
           
           optional arguments:
           -h, --help  show this help message and exit
           -C C        Category search pattern
           -D D        Destination category name

        """
        # Set up positional and optional arguments
        parser = argparse.ArgumentParser()
        # This is the target string for the category
        parser.add_argument("-C", help="Category search pattern",
                            action='store', required=True)
        parser.add_argument("-D", help="Destination category name",
                            action='store')
        
        # Parse the command line as supplied
        args = parser.parse_args()
        # Set up the member value according to the command line options
        self._cat = args.C
        self._dest = args.D
        if self._cat is not None:
            if self._dest is None:
                self._dest = self._cat
        


    def process(self):
        """The execution function for the app.

        The basic flow:

        #. A destination subdirectory is created, unless it already exists. 

        #. The app then examines all the files in the current directory.
        
        #. If the specfied category search substring is found in the filename, that file is moved into the subdirectory.
        """
        mkdir_p(self._dest)
        # define the path
        currentDirectory = pathlib.Path('.')

        for currentFile in currentDirectory.iterdir():
            if self._cat in currentFile.stem:
                if currentFile.is_file() is True:
                    target = pathlib.Path(self._dest, currentFile)
                    print(target)
                    currentFile = currentFile.replace(target)
                    print(currentFile)



"""
Abstract base class for all Process-based generator algorithms.

Copyright 2020 Thomas Jackson Park & Jeremy Pavier

"""
import os
import math
import random
from pydub import AudioSegment
from GenerIter.selector import Selector
from GenerIter.source import WavSource
from GenerIter.util import debug, jStr, mkdir_p

class Process():
    """This is the abstract base class from which all other processors are derived.

    As such it implements the core interface as well as several important generic helper services which can simplify the derived algorithm implementations.
    """
    # Currently only WAV is supported, but this list is expected to grow
    SUPPORTED_FORMATS = ["wav"]

    TSHIRT = {
        "s" : 180,
        "m" : 300,
        "l" : 480,
        "xl" : 780,
        "xxl" : 1260,
        "xxxl" : 2640
    }

    def __init__(self, prefix=None):
        self._config = None
        self._inventory = None
        self._destination = None
        self._format = "wav"
        self._content = []
        self._prefix = prefix
        self._tsize = "m"
        debug('Process()')

    def configure(self, inventory, configuration, destination, forrmat, tsize):
        self._config = configuration
        self._inventory = inventory
        self._destination = destination
        self._format = forrmat
        self._tsize = tsize

    def default(self):
        debug('No-op default processing logic')
        debug(type(self))

    def declick(self, segment, value=10):
        """This is a helper function with which a fade/rise can be applied to each end of an AudioSegment to reduce the potential for 
        'clicking' when they are connected end-to-end.

        Can also be used, with longer fade times, if large track fades are required.

        Args:
            value (int) : number of milliseconds across which the segment will be faded from full gain to zero (typically 10 seems to work well).

        Returns:
            AudioSegment
        """
        segment = segment.fade_in(value)
        segment = segment.fade_out(value)
        return segment

    def deamplify(self, segment, limits):
        diminish = random.randrange(limits[0], limits[1])
        segment = segment - diminish
        return segment

    #def getsegment(self, sample, limits, fade):
    #    retval = AudioSegment.from_wav(sample)
    #    retval = self.deamplify(retval, limits)
    #    retval = self.declick(retval, fade)
    #    return retval

    def getsegment(self, sample, muted, fade):
        retval = AudioSegment.from_wav(sample)
        if muted > 0:
            retval = retval - muted
        if fade > 0:
            retval = self.declick(retval, fade)
        return retval


    def intwidth(self, value):
        retval = 1 + int(math.log10(value))
        return retval

    def supported(self, value):
        return value in self.SUPPORTED_FORMATS

    def threshold(self):
        if self._tsize in self.TSHIRT:
            target = self.TSHIRT[self._tsize]
        else:
            target = self.TSHIRT["m"]
        lbnd = (9 * target) // 10
        ubnd = (11 * target) // 10
        retval = random.randrange(lbnd, ubnd)
        return retval

    def padtolength(self, segment, length, fader, front=False):
        quiet = length - len(segment)
        if quiet <= 0:
            retval = segment[:length]
            retval = self.declick(retval, fader)
        else:
            pad = AudioSegment.silent(duration=quiet,
                                      frame_rate=segment.frame_rate)
            if front is False:
                retval = segment + pad
            else:
                retval = pad + segment
        return retval

    def bracket(self, segment, frontmult=1.0, backmult=1.0):
        frunt = random.randrange(int(frontmult * len(segment)))
        bak = random.randrange(int(backmult * len(segment)))
        front = AudioSegment.silent(duration=frunt, frame_rate=segment.frame_rate)
        back = AudioSegment.silent(duration=bak, frame_rate=segment.frame_rate)
        retval = front + segment + back
        return retval
        
        

    def write(self, algorithm, counter, source):
        # How many times do you want this to run?
        iterations = int(self._config["tracks"])
        # Need to be able to pad the correct number of zeroes
        digits = self.intwidth(iterations)
        # What's the base root name of the outputs?
        base = type(self).__name__
        # Set the correct sub directory and ensure it exists
        destdir = os.path.join(self._destination, base)
        mkdir_p(destdir)
        # Zero pad the counter string
        ctr = str(counter).zfill(digits)
        track = f"{base}_{algorithm}_{ctr}"
        debug(f"Track: {track}")

        # Create the output file name with zero padded counter value
        if self.supported(self._format) is True:
            filename = "{0}.{1}".format(track, self._format)
            # Set up the output path
            dest = os.path.join(destdir, filename)
            print(f"\t{dest}")
            # Write it out
            source.export(dest, format=self._format)
            src = WavSource(dpath=dest, dexist=True)
            self._inventory.insert(src)
        else:
            print("Unsupported output format : {0}".format(self._format))
